import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from .cmaps import redblue_cmap

def cor_heatmap(x, figsize=[7, 7], gridspec_kw=None, vmin=-1, vmax=1, cmap=redblue_cmap, linewidths=0.01,
    linecolor='white', cbar=True, cbar_kws=None, square=True, xticklabels=None, yticklabels=None, mask=None, upper=True,
    xtick_labelsize=8, ytick_rotation=0, ytick_labelsize=8, cbar_labelsize=8, fname=None, quality=100, transparent=True,
    bbox_inches='tight', pad_inches=0.1):
    num_params = x.shape[0]

    gridspec_kw = gridspec_kw or {'height_ratios': (0.93, .02), 'hspace': .03}

    cbar_kws = cbar_kws or {'orientation': 'horizontal'}
    xticklabels = xticklabels or range(1, num_params+1)
    yticklabels = yticklabels or range(1, num_params+1)
    if (mask is None) and upper:
        mask = np.tril(np.ones_like(x, dtype=np.bool))

    fig, (ax, cbar_ax) = plt.subplots(2, figsize=figsize, gridspec_kw=gridspec_kw)

    ax = sns.heatmap(
        x,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        linewidths=linewidths,
        linecolor=linecolor,
        cbar=cbar,
        cbar_kws=cbar_kws,
        cbar_ax=cbar_ax,
        square=square,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        mask=mask,
        ax=ax
    )
    
    ax.xaxis.tick_top()
    ax.yaxis.tick_right()

    ax.set_xticklabels(ax.get_xticklabels(), fontsize=xtick_labelsize)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=ytick_rotation, fontsize=ytick_labelsize)

    ax.collections[0].colorbar.ax.tick_params(labelsize=cbar_labelsize)

    if fname is not None:
        plt.savefig(fname, quality=quality, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)



from functools import reduce
from operator import add

from .cor import cor

# x is a numpy array of 3 dimensions, (chain, MC iteration, parameter)
def mean_cor(x):
    return reduce(add, map(lambda m: cor(m), x))/len(x)



import matplotlib.colors as mcolors
import matplotlib.pyplot as plt

from .cmaps import redblue_cmap

def hist2d(x, y, figsize=[6.4, 4.8], bins=20, ranges=None, density=False,
    cmap=redblue_cmap, norm=mcolors.Normalize(), vmin=None, vmax=None, alpha=None, xlabel=None, ylabel=None,
    axes_labelsize=14, title=None, axes_titlesize=14, xticks=None, xticklabels=None, xtick_labelsize=12, yticks=None,
    yticklabels=None, ytick_labelsize=12, cbar=False, cbar_kws=None, cbar_labelsize=8, fname=None, quality=100,
    transparent=True, bbox_inches='tight', pad_inches=0.1):
    cbar_kws = cbar_kws or {'orientation': 'vertical'}

    plt.figure(figsize=figsize)

    plt.rcParams['axes.labelsize'] = axes_labelsize
    plt.rcParams['axes.titlesize'] = axes_titlesize
    plt.rcParams['xtick.labelsize'] = xtick_labelsize
    plt.rcParams['ytick.labelsize'] = ytick_labelsize

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if title is not None:
        plt.title(title)

    if (xticks is not None):
        plt.xticks(ticks=xticks, labels=xticklabels or [str(xtick) for xtick in xticks])
    if (yticks is not None):
        plt.yticks(ticks=yticks, labels=yticklabels or [str(ytick) for ytick in yticks])

    out_h, out_xedges, out_yedges, out_image = plt.hist2d(
        x, y, bins=bins, range=ranges, density=density, cmap=cmap, norm=norm, vmin=vmin, vmax=vmax, alpha=alpha
    )

    if cbar:
        colorbar = plt.colorbar(**cbar_kws)
        colorbar.ax.tick_params(labelsize=cbar_labelsize)

    if fname is not None:
        plt.savefig(fname, quality=quality, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)

    return out_h, out_xedges, out_yedges, out_image



# See lemma 2 of section 4.1 in https://www.tandfonline.com/doi/abs/10.1080/10618600.1998.10474787

import numpy as np

from .mc_cov import mc_cov

# x is a numpy array of 3 dimensions, (chain, MC iteration, parameter)
def multi_rhat(x, method='inse', adjust=False, b=None, r=3):
    num_chains, num_iters, num_pars = x.shape

    w = np.zeros([num_pars, num_pars])
    for i in range(num_chains):
        w = w + mc_cov(x[i], method=method, adjust=adjust, b=b, r=r)
    w = w / num_chains

    b = np.cov(np.apply_along_axis(np.mean, 1, x), rowvar=False)

    rhat = max(np.linalg.eig(np.matmul(np.linalg.inv(w), b))[0])
    rhat = ((num_iters - 1) / num_iters) + ((num_chains + 1) / num_chains) * rhat

    return rhat, w, b



import numpy as np

def is_pos_def(x):
    if np.array_equal(x, x.transpose()):
        try:
            np.linalg.cholesky(x)
            return True
        except np.linalg.LinAlgError:
            return False
    else:
        return False



import numpy as np

from kanga.emulators import optimal_ar_batch_size

from .bm_cov import bm_cov

def bm_mc_cov(x, b=None, r=3):
    if b is None:
        b = optimal_ar_batch_size(x)

    if ((r < 0) or (r > 5)):
        raise ValueError(' r={}. Pick a value of r in between 1 and 5'.format(r))

    if (b == 1 and r != 1):
        raise ValueError('r={} and b={}. If b=1, then r should be set to 1 too'.format(r, b))

    br = int(np.floor(b / r))
    if (br < 1):
        raise ValueError('floor(b/r)={}<1. Increase b or decrease r'.format(br))

    cov_mat = bm_cov(x, b)

    if (r > 1):
        # Lugsail lag window, see p. 4, eq. 2 in https://arxiv.org/pdf/1809.04541.pdf
        lug_cov_mat = 2 * cov_mat - bm_cov(x, br)

        if (lug_cov_mat.diagonal().prod() > 0):
            cov_mat = lug_cov_mat

    return cov_mat



import matplotlib.pyplot as plt
import numpy as np

from math import sqrt
from scipy.stats import norm

def acf(data, maxlag=40, normed=True, usevlines=True, figsize=[6.4, 4.8], linewidth=None, linestyle=None, marker=None,
    markersize=None, color=None, ylim=None, margins=0., title=None, xlabel=None, ylabel=None, axes_labelsize=14,
    axes_titlesize=14, xticks=None, xticklabels=None, xtick_labelsize=12, yticks=None, yticklabels=None, ytick_labelsize=12,
    hlines=True, a=0.05, hline_linewidth=1, hline_linestyle='dashed', hline_color='#1f77b4', legend=False,
    legend_labels=None, legend_loc='best', legend_ncol=None, legend_fontsize=12, fname=None, quality=100, transparent=True,
    bbox_inches='tight', pad_inches=0.1):
    num_datasets = 1 if np.isscalar(data[0]) else len(data[0])
    num_samples = len(data)

    linewidth = linewidth or [1.5 for _ in range(num_datasets)]
    linestyle = linestyle or ['solid' for _ in range(num_datasets)]
    color = color or ['black' for _ in range(num_datasets)]
    if not usevlines:
        marker = marker or ['o' for _ in range(num_datasets)]
        markersize = markersize or [4 for _ in range(num_datasets)]

    if legend:
        legend_labels = legend_labels or [str(i+1) for i in range(num_datasets)]

    plt.figure(figsize=figsize)

    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    plt.xlim([0, maxlag])

    if ylim is not None:
        plt.ylim(ylim)

    plt.margins(margins)

    plt.rcParams['axes.labelsize'] = axes_labelsize
    plt.rcParams['axes.titlesize'] = axes_titlesize
    plt.rcParams['xtick.labelsize'] = xtick_labelsize
    plt.rcParams['ytick.labelsize'] = ytick_labelsize
    if legend:
        plt.rcParams['legend.fontsize'] = legend_fontsize

    if title is not None:
        plt.title(title)

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if (xticks is not None):
        plt.xticks(ticks=xticks, labels=xticklabels or [str(xtick) for xtick in xticks])
    if (yticks is not None):
        plt.yticks(ticks=yticks, labels=yticklabels or [str(ytick) for ytick in yticks])

    out_lags = [None] * num_datasets
    out_c = [None] * num_datasets
    out_line = [None] * num_datasets
    out_b = [None] * num_datasets

    for i in range(num_datasets):
        opt_args = {
            'maxlags': maxlag,
            'normed': normed,
            'usevlines': usevlines,
            'linewidth': linewidth[i],
            'linestyle': linestyle[i],
            'color': color[i]
        }
        if not usevlines:
            opt_args['marker'] = marker[i]
            opt_args['markersize'] = markersize[i]
        if legend:
            opt_args['label'] = legend_labels[i]

        out_lags[i], out_c[i], out_line[i], out_b[i] = plt.acorr(data if (num_datasets == 1) else data[:, i], **opt_args)

    # https://stats.stackexchange.com/questions/211628/how-is-the-confidence-interval-calculated-for-the-acf-function
    # https://stackoverflow.com/questions/24695174/python-equivalent-of-qnorm-qf-and-qchi2-of-r
    if hlines:
        upper_bound = norm.ppf(1-a/2)/sqrt(num_samples)
        plt.hlines(
            upper_bound, xmin=0, xmax=maxlag, linewidth=hline_linewidth, linestyle=hline_linestyle, color=hline_color
        )
        plt.hlines(
            -upper_bound, xmin=0, xmax=maxlag, linewidth=hline_linewidth, linestyle=hline_linestyle, color=hline_color
        )

    if legend:
        plt.legend(labels=legend_labels, loc=legend_loc, ncol=legend_ncol)

    if fname is not None:
        plt.savefig(fname, quality=quality, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)

    return out_lags, out_c, out_line, out_b



import matplotlib.pyplot as plt
import numpy as np

from cycler import cycler

def trace(y, x=None, figsize=[8, 4], linewidth=None, color=None, ylim=None, margins=0., title=None, xlabel=None, ylabel=None,
    axes_labelsize=14, axes_titlesize=14, xticks=None, xticklabels=None, xtick_labelsize=12, yticks=None, yticklabels=None,
    ytick_labelsize=12, legend=False, legend_labels=None, legend_loc='best', legend_ncol=None, legend_fontsize=12,
    fname=None, quality=100, transparent=True, bbox_inches='tight', pad_inches=0.1):
    num_traces = 1 if np.isscalar(y[0]) else len(y[0])

    linewidth = linewidth or [1.5 for _ in range(num_traces)]
    color = color or ['#1f77b4' for _ in range(num_traces)]
    prop_cycle =  cycler(linewidth=linewidth) + cycler(color=color)

    if legend:
        legend_labels = legend_labels or [str(i+1) for i in range(num_traces)]

    plt.figure(figsize=figsize)

    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    plt.gca().set_prop_cycle(prop_cycle)

    if ylim is not None:
        plt.ylim(ylim)

    plt.margins(margins)

    plt.rcParams['axes.labelsize'] = axes_labelsize
    plt.rcParams['axes.titlesize'] = axes_titlesize
    plt.rcParams['xtick.labelsize'] = xtick_labelsize
    plt.rcParams['ytick.labelsize'] = ytick_labelsize
    if legend:
        plt.rcParams['legend.fontsize'] = legend_fontsize

    if title is not None:
        plt.title(title)

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if (xticks is not None):
        plt.xticks(ticks=xticks, labels=xticklabels or [str(xtick) for xtick in xticks])
    if (yticks is not None):
        plt.yticks(ticks=yticks, labels=yticklabels or [str(ytick) for ytick in yticks])

    handles = plt.plot(x, y) if (x is not None) else plt.plot(y)

    if legend:
        plt.legend(handles=handles, labels=legend_labels, loc=legend_loc, ncol=legend_ncol)

    if fname is not None:
        plt.savefig(fname, quality=quality, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)



# https://www.sciencedirect.com/science/article/pii/S0047259X16301877
# https://arxiv.org/pdf/1706.00853.pdf
# The notation follows the implementation of insec.cpp of the mcmcse R package

import numpy as np

from kanga.linalg import is_pos_def

def inse_mc_cov(x, adjust=False):
    x = x - x.mean(0)
    
    n, p = x.shape
    
    ub = int(np.floor(n / 2))
    sn = ub

    if adjust:
        Gamadj = np.zeros([p, p])

    for m in range(ub):
        gam0 = np.zeros([p, p])
        gam1 = np.zeros([p, p])

        for i in range(n - 2 * m):
            gam0 = gam0 + np.outer(x[i, :], x[i + 2 * m, :])
        gam0 = gam0 / n

        for i in range(n - 2 * m - 1):
            gam1 = gam1 + np.outer(x[i, :], x[i + 2 * m + 1, :])
        gam1 = gam1 / n

        Gam = gam0 + gam1
        Gam = (Gam + Gam.transpose()) / 2

        if m == 0:
            Sig = -gam0 + 2 * Gam
        else:
            Sig = Sig + 2 * Gam

        if is_pos_def(Sig):
            sn = m
            break

    if sn > (ub-1):
        raise RuntimeError('Not enough samples')

    last_dtm = np.linalg.det(Sig)

    for m in range(sn + 1, ub):
        gam0 = np.zeros([p, p])
        gam1 = np.zeros([p, p])

        for i in range(n - 2 * m):
            gam0 = gam0 + np.outer(x[i, :], x[i + 2 * m, :])
        gam0 = gam0 / n

        for i in range(n - 2 * m - 1):
            gam1 = gam1 + np.outer(x[i, :], x[i + 2 * m + 1, :])
        gam1 = gam1 / n

        Gam = gam0 + gam1
        Gam = (Gam + Gam.transpose()) / 2

        Sig1 = Sig + 2 * Gam

        current_dtm = np.linalg.det(Sig1)

        if current_dtm <= last_dtm:
            break

        Sig = Sig1.copy()

        last_dtm = current_dtm

        if adjust:
            eigenvals, eigenvecs = np.linalg.eig(Gam)
            eigenvals[eigenvals > 0] = 0
            Gamadj = Gamadj - eigenvecs @ np.diag(eigenvals) @ np.linalg.inv(eigenvecs)

    if adjust:
        Sig = Sig + 2 * Gamadj

    return Sig



import numpy as np

from .bm_mc_cov import bm_mc_cov
from .inse_mc_cov import inse_mc_cov

def mc_cov(x, method='inse', adjust=False, b=None, r=3):
    if method == 'inse':
        return inse_mc_cov(x, adjust=adjust)
    elif method == 'bm':
        return bm_mc_cov(x, b=b, r=r)
    elif method == 'iid':
        return np.cov(x, rowvar=False)
    else:
        raise ValueError('The method can be inse, bm or iid, {} was given'.format(method))



from .batch_mean import batch_mean
from .batch_var import batch_var
from .bm_cov import bm_cov
from .cor import cor
from .bm_mc_cov import bm_mc_cov
from .inse_mc_cov import inse_mc_cov
from .mc_cov import mc_cov
from .mean_cor import mean_cor
from .multi_ess import multi_ess
from .multi_rhat import multi_rhat
from .running_mean import running_mean



import numpy as np

from statsmodels.tsa.ar_model import ar_select_order

def fit_ar(x):
    n = len(x)

    ar_selection = ar_select_order(
        x,
        min(int(np.floor(n)), int(np.floor(10 * np.log10(n)))),
        ic='aic',
        trend='n',
        seasonal=False
    )

    order = ar_selection.ar_lags[-1]

    model_fit = ar_selection.model.fit()

    return model_fit.params, model_fit.sigma2, order



import matplotlib.pyplot as plt
import numpy as np

def hist(data, figsize=[6.4, 4.8], bins=20, xrange=None, density=False, linewidth=1.5, color=None, edgecolor='white',
    alpha=None, ylim=None, margins=0., xlabel=None, ylabel=None, axes_labelsize=14, title=None, axes_titlesize=14,
    xticks=None, xticklabels=None, xtick_labelsize=12, yticks=None, yticklabels=None, ytick_labelsize=12, legend=False,
    legend_labels=None, legend_loc='best', legend_ncol=None, legend_fontsize=12, fname=None, quality=100, transparent=True,
    bbox_inches='tight', pad_inches=0.1):
    if legend:
        num_datasets = 1 if np.isscalar(data[0]) else len(data[0])
        legend_labels = legend_labels or [str(i+1) for i in range(num_datasets)]

    plt.figure(figsize=figsize)

    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    plt.rcParams['axes.labelsize'] = axes_labelsize
    plt.rcParams['axes.titlesize'] = axes_titlesize
    plt.rcParams['xtick.labelsize'] = xtick_labelsize
    plt.rcParams['ytick.labelsize'] = ytick_labelsize
    if legend:
        plt.rcParams['legend.fontsize'] = legend_fontsize

    if ylim is not None:
        plt.ylim(ylim)

    plt.margins(margins)

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if title is not None:
        plt.title(title)

    if (xticks is not None):
        plt.xticks(ticks=xticks, labels=xticklabels or [str(xtick) for xtick in xticks])
    if (yticks is not None):
        plt.yticks(ticks=yticks, labels=yticklabels or [str(ytick) for ytick in yticks])

    hist_opt_args ={
        'bins': bins,
        'range': xrange,
        'density': density,
        'linewidth': linewidth,
        'edgecolor': edgecolor,
        'alpha': alpha
    }
    if color is not None:
        hist_opt_args['color'] = color
    if legend:
        hist_opt_args['label'] = legend_labels

    out_n, out_bins, out_patches = plt.hist(data, **hist_opt_args)

    if legend:
        plt.legend(loc=legend_loc, ncol=legend_ncol)

    if fname is not None:
        plt.savefig(fname, quality=quality, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)

    return out_n, out_bins, out_patches



from codecs import open
from os import path
from setuptools import find_packages, setup

from kanga import __version__

url = 'https://github.com/papamarkou/kanga'

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='kanga',
    version=__version__,
    description='MCMC diagnostics',
    long_description=long_description,
    url=url,
    download_url='{0}/archive/v{1}.tar.gz'.format(url, __version__),
    packages=find_packages(),
    license='MIT',
    author='Theodore Papamarkou',
    author_email='theodore.papamarkou@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords=['Bayesian', 'diagnostics', 'Markov chains', 'MCMC', 'Monte Carlo'],
    install_requires=['numpy', 'scipy', 'statsmodels>=0.11.1', 'matplotlib', 'seaborn']
)
